@@ -1,5 +1,3 @@
-#
-#
 
 package PSA::Acceptor;
 
@@ -9,10 +7,15 @@ PSA::Acceptor - Base class for back-ends to accept PSA requests
 
 =head1 SYNOPSIS
 
-my $acceptor = PSA::Acceptor::AutoCGI->new();
+ my $acceptor = PSA::Acceptor::AutoCGI->new();
 
 =head1 DESCRIPTION
 
+Normally, you create an instance of a sub-class of this module.
+Currently the only acceptors that has been written is
+L<PSA::Acceptor::AutoCGI>.  There should be a L<PSA::Acceptor::TCP>,
+as well, for convenience.
+
 =cut
 
 use base qw(Class::Tangram);
@@ -59,8 +62,14 @@ sub stale {
     return ( $self->{hit_count} >= $self->{max_age} );
 }
 
+1;
+
+__END__
+
 =back
 
-=cut
+=head1 SEE ALSO
 
-1;
+L<PSA>, L<PSA::Acceptor::AutoCGI>, L<PSA::Acceptor::Dummy>
+
+=cut
